#include <stdlib.h>

#include "hruby_sim.h"

/**
 *  The lists used in HDLRuby simulation, to be used with C code generated by
 *  hruby_low2c. 
 *  */


/* The pool of list elements to reduce number of memory allocations. */
static ListS pool_elements = { NULL, NULL };

/** Get a list element for containing some data.
 *  @param data the data of the element
 *  @return the resulting element */
Elem get_element(void* data) {
    Elem elem; /* The resulting element. */
    /* Is the pool empty? */
    if (empty_list(&pool_elements)) {
        /* Yes, allocates a new element. */
        elem = malloc(sizeof(ElemS));
        elem->data = data;
        elem->next = NULL;
    } else {
        /* No, get the element from the pool. */
        elem = pool_elements.head;
        elem->data = data;
        elem->next = NULL;
        pool_elements.head = pool_elements.head->next;
        if (pool_elements.tail == elem)
            pool_elements.tail = NULL;
    }
    return elem;
}

/** Delete an element (it will return to the pool).
  *  @param elem the element to delete */
void delete_element(Elem elem) {
    elem->data = NULL;
    add_list(&pool_elements,elem);
}


/** Builds a list.
 *  @param list the place where to build the list
 *  @return the resulting list */
List build_list(List list) {
    list->head = list->tail = NULL;
    return list;
}

/** Clears a list.
 *  @param list the list to clear */
void clear_list(List list) {
    list->head = list->tail = NULL;
}

/** Adds an element to the tail of a list.
 *  @param list the list to add the element in
 *  @param elem the element to add in the list */
void add_list(List list, Elem elem) {
    /* Ensures the element is unlinked. */
    elem->next = NULL;
    /* Is the list empty? */
    if (empty_list(list)) {
        /* Yes, the element is the head and the tail. */
        list->head = list->tail = elem;
    } else {
        /* No, simply add to the tail. */
        list->tail->next = elem;
        list->tail = elem;
    }
}

/** Remove the head of the list.
 *  @param list the list to remove the head from
 *  @return the removed element */
Elem remove_list(List list) {
    /* Is the list empty? */
    if (empty_list(list)) {
        /* Yes, no element to remove. */
        return NULL;
    } else {
        /* No, remove the head. */
        Elem elem = list->head;
        list->head = list->head->next;
        if (list->tail == elem) {
            list->tail = NULL;
        }
        return elem;
    }
}
